/*

    File: fat.c

    Copyright (C) 1998-2000  Christophe Grenier <grenier@nef.esiea.fr>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#include <stdio.h>
#include "types.h"
#include "common.h"
#include "fat.h"
#include "lang.h"
#include "hdaccess.h"
#include "fnctdsk.h"
#include "testdisk.h"
extern t_param_disk disk_car;
uint sector_size(struct fat_boot_sector *fat_header);
uint dir_entries(struct fat_boot_sector *fat_header);
uint sectors(struct fat_boot_sector *fat_header);

int check_FAT(t_diskext partition,int debug)
{
  t_sector buffer;
#ifdef DEV
  recover_FAT(partition,debug);
#endif
  if(hd_read2(1, &buffer, partition))
  {
    doprintf(m_CHKFAT_RERR); return 1;
  }
  return check_FAT_aux((struct fat_boot_sector *)buffer,partition,debug);
}

#ifdef DEV
/* The idea is to be able to rebuild boot sector and first FAT
 * when they are been erased by CIH virus */
dword search_fat16(t_diskext partition,int debug)
{
  dword i;
  uchar buffer[0x200];
  t_disk pos_fat=(t_disk)MALLOC(sizeof(*pos_fat));
//  doprintf("\nsearch_fat16");
  pos_fat->disk=partition->disk;
  set_pos(pos_fat,get_LBA_part(partition));
//  doprintf("\nsearch_fat16 test %u %u\n",partition->ssector,pos_fat->sector);
  for(i=0;i<65535;i++)
  {
//    aff_part(0,partition);
    if(hd_read(1, &buffer, pos_fat))
    {
      FREE(pos_fat);
      return 0;
    }
//    dump(buffer,0x200);
    if(buffer[0]==0xF8 && buffer[1]==0xFF)
    {
      FREE(pos_fat);
      if(buffer[2]!=0xFF)
	partition->part_type=P_12FAT;
      else
      if(buffer[3]!=0xFF)
	partition->part_type=P_16FAT;
      else
	partition->part_type=P_32FAT_1;
      return i;
    }
    disk_apres(pos_fat);
  }
  FREE(pos_fat);
  return 0;
}

int recover_FAT(t_diskext partition,int debug)
{
  dword start_fat1,start_fat2,start_rootdir,start_data,no_of_cluster;
  t_diskext partition2=(t_diskext)MALLOC(sizeof(*partition2));
  partition2->disk=partition->disk;
  set_posexts(partition2,get_SR_part(partition));	/* Don't skip MBR */
  start_fat1=search_fat16(partition2,debug);
  if(start_fat1<=0)
  {
    doprintf("Unable to found first FAT");
    FREE(partition2);
    return 1;
  }
  /* start_fat1 must be 1 */
  /* skip first FAT1 sector */
  set_posexts(partition2,get_SR_part(partition)+start_fat1+1);
  start_fat2=search_fat16(partition2,debug);
  aff_part(0,partition);
  if(start_fat2==0)
  {
      FREE(partition2);
      return 1;
  }
  if(start_fat2==0)
  {
    start_fat2=start_fat1;
    start_fat1=0;
  }
  else
    start_fat2+=start_fat1+1;
  doprintf("\nrecover_FAT  FAT 1: %u - FAT 2: %u\n", start_fat1,start_fat2);
  FREE(partition2);
  return 1; /* Echec */
}
#endif

int check_FAT_aux(struct fat_boot_sector *fat_header, t_diskext partition,int debug)
{
  dword start_fat1,start_fat2,start_rootdir,start_data,no_of_cluster,fat_length,fat_length_calc,part_size,end_data;
  if(fat_header->marker!=0xAA55)
  { doprintf("check_FAT : " m_BAD_NMARK); return 1; }
  if(!((fat_header->ignored[0]==0xeb&&fat_header->ignored[2]==0x90)||fat_header->ignored[0]==0xe9))
  { doprintf(m_CHKFAT_BAD_JUMP); return 1; }
  if(sector_size(fat_header)!=512)
  { doprintf(m_CHKFAT_BYTESECT); return 1; }
  switch(fat_header->cluster_size)
  { /* should depend of FAT12/16/32 */
    case 1:case 2:case 4:case 8:case 16:case 32:case 64:case 128: break;
    default: doprintf(m_CHKFAT_SECT_CLUSTER); return 1;
  }
  if(fat_header->secs_track!=disk_car->sector)
  { doprintf(m_CHKFAT_SECTTRACK); return 1; }
  if(fat_header->heads!=disk_car->head+1)
  { doprintf(m_CHKFAT_SIDE); return 1; }
  if(fat_header->fats!=2)
  { doprintf(m_CHKFAT_NBRFAT); return 1; }
  if(dir_entries(fat_header)%16)        /* Improved it */
  { doprintf(m_CHKFAT_ENTRY); return 1; }
  if(fat_header->media!=0xF8)
  { doprintf(m_CHKFAT_MEDIA); return 1; }
  switch(partition->part_type)
  {
    case P_12FAT:
	  part_size=sectors(fat_header);
      if(part_size!=partition->part_size)
      { doprintf(m_CHKFAT_SIZE); return 1; }
      if(fat_header->reserved!=1)
      { doprintf(m_CHKFAT_RESERV); return 1; }
      if(dir_entries(fat_header)==0)
      { doprintf(m_CHKFAT_ENTRY); return 1; }
      if(fat_header->fat_length>256)
      { doprintf(m_CHKFAT_SECTPFAT); return 1; }
	  fat_length=fat_header->fat_length;
	  start_fat1=fat_header->reserved;
      start_fat2=start_fat1+fat_length;
      start_rootdir=start_fat2+fat_length;
      start_data=start_rootdir+(dir_entries(fat_header)-1)/16+1;
	  no_of_cluster=(part_size-start_data)/fat_header->cluster_size;
	  fat_length_calc=((no_of_cluster+1)*1.5/512)+1;
	  break;
    case P_16FAT:
    case P_16FATH:
    case P_16FATBD:
    case P_16FATBDH:
	  part_size=fat_header->total_sect;
      if((sectors(fat_header)!=0)||(part_size!=partition->part_size))
      { doprintf(m_CHKFAT_SIZE); return 1; }
   	  fat_length=fat_header->fat_length;
	  start_fat1=fat_header->reserved;
      start_fat2=start_fat1+fat_length;
      start_rootdir=start_fat2+fat_length;
      start_data=start_rootdir+(dir_entries(fat_header)-1)/16+1;
	  no_of_cluster=(fat_header->total_sect-start_data)/fat_header->cluster_size;
	  fat_length_calc=((no_of_cluster+1)*2/512)+1;
	  break;
    case P_32FAT_1:
    case P_32FAT_2:
    case P_32FAT_1H:
    case P_32FAT_2H:
	  part_size=fat_header->total_sect;
      if((sectors(fat_header)!=0)||
	  (part_size!=partition->part_size))
      { doprintf(m_CHKFAT_SIZE); return 1; }
      fat_length=fat_header->fat_length?fat_header->fat_length:fat_header->fat32_length;
      start_fat1=fat_header->reserved;
      start_fat2=start_fat1+fat_length;
      start_data=start_fat2+fat_length;
      start_rootdir=start_data+fat_header->root_cluster*fat_header->cluster_size;
      no_of_cluster=(part_size-start_data)/fat_header->cluster_size;
	  fat_length_calc=((no_of_cluster+1)*4/512)+1;
	  break;
	default:
	  return 1;
  }
  end_data=start_data+no_of_cluster*fat_header->cluster_size;
  if(debug)
  {
	doprintf("\nFAT1 : %u-%u",start_fat1,start_fat1+fat_length-1);
	doprintf("\nFAT2 : %u-%u",start_fat2,start_fat2+fat_length-1);
	doprintf("\nstart_rootdir : %u",start_rootdir);
	doprintf("\nData : %u-%u",start_data,end_data);
	doprintf("\nsectors : %u",part_size);
	doprintf("\ncluster_size : %u",fat_header->cluster_size);
	doprintf("\nno_of_cluster : %u (2 - %u)", no_of_cluster,no_of_cluster+1);
	doprintf("\nfat_length %u calculated %u\n",fat_length,fat_length_calc);
  }
  if(fat_length<fat_length_calc)
  { doprintf(m_CHKFAT_SECTPFAT); return 1; }
  return comp_FAT(partition,fat_length,fat_header->reserved);
}

int comp_FAT(t_diskext partition,dword taille,const dword sect_res)
{
  uint read_size;
  uint i;
  uchar buffer[NBR_SECT*0x200],buffer2[NBR_SECT*0x200];
  t_disk pos_fat=(t_disk)MALLOC(sizeof(*pos_fat));
  t_disk pos_fat2=(t_disk)MALLOC(sizeof(*pos_fat2));
  fflush(stdout);
  pos_fat->disk=partition->disk;
  pos_fat2->disk=partition->disk;
  set_pos(pos_fat,get_LBA_part(partition)+sect_res);
  set_pos(pos_fat2,get_LBA_part(partition)+sect_res+taille);
  while(taille)
  {
    read_size=taille>NBR_SECT?NBR_SECT:taille;
    taille-=read_size;
    if(hd_read(read_size, &buffer, pos_fat))
    { doprintf(m_CHKFAT_RERR);
      FREE(pos_fat);
      FREE(pos_fat2);
      return 1;}
    if(hd_read(read_size, &buffer2, pos_fat2))
    { doprintf(m_CHKFAT_RERR); 
      FREE(pos_fat);
      FREE(pos_fat2);
      return 1;}
    if(memcmp(buffer,buffer2,0x200*read_size)!=0)
    { doprintf("\nFAT differs"); 
      FREE(pos_fat);
      FREE(pos_fat2);
      return 1;
    }
    for(i=0;i<read_size;i++)
    {
      disk_apres(pos_fat);
      disk_apres(pos_fat2);
    }
  }
  FREE(pos_fat);
  FREE(pos_fat2);
  return 0;
}

uint sector_size(struct fat_boot_sector *fat_header)
{ return (fat_header->sector_size[1]<<8)+fat_header->sector_size[0]; }

uint dir_entries(struct fat_boot_sector *fat_header)
{ return (fat_header->dir_entries[1]<<8)+fat_header->dir_entries[0]; }

uint sectors(struct fat_boot_sector *fat_header)
{ return (fat_header->sectors[1]<<8)+fat_header->sectors[0]; }

int analyse_FAT(struct fat_boot_sector*fat_header, t_diskext partition,int debug, int dump_ind)
{
  char*buffer=(char*)fat_header;
  if(fat_header->marker==0xAA55)
  {
	if(debug||dump_ind)
	{
	  doprintf("\nMarker (0xAA55) at %u/%u/%u\n", partition->scylinder,partition->shead,partition->ssector);
	}
	if(dump_ind)
	  dump(buffer,0x200);
    if(memcmp(buffer+FAT_NAME1,"FAT12",5)==0) /* 2 Mo max */
    {
      partition->part_size=sectors(fat_header);
      partition->part_type=P_12FAT;
      return 0;
    }
    else if(memcmp(buffer+FAT_NAME1,"FAT16",5)==0)
    {
      if(sectors(fat_header)!=0)
      {
	partition->part_size=sectors(fat_header);
	partition->part_type=P_16FAT;
	return 0;
      }
      partition->part_size=fat_header->total_sect;
      partition->part_type=P_16FATBD;
    }
    else if(memcmp(buffer+FAT_NAME2,"FAT32",5)==0)
    {
      if(sectors(fat_header)!=0)
	partition->part_size=sectors(fat_header);
      else
	partition->part_size=fat_header->total_sect;
      partition->part_type=P_32FAT_1;
      return 0;
    }
    else if(memcmp(buffer+OS2_NAME,"IBM",3)==0)
    {   /* D'apres une analyse de OS2 sur systeme FAT...
	FAT_NAME1=FAT
	*/
      if(sectors(fat_header)!=0)
	partition->part_size=sectors(fat_header);
      else
	partition->part_size=fat_header->total_sect;
      partition->part_type=P_HPFS;
      return 0;
    }
    else if(memcmp(buffer+FAT_NAME1,"FAT",3)==0)
    {   /* Attention a detecter apres OS2 */
      partition->part_size=(disk_car->head+1) * disk_car->sector; /* 1 cylinder */
      partition->part_type=P_OS2MB;
      return 0;
    }
    else if(memcmp(buffer+NTFS_NAME,"NTFS",4)==0)
    {
      partition->part_size=NT_sect(buffer)+1;
      partition->part_type=P_NTFS;
      return 0;
    }
  }     /* fin marqueur de fin :)) */
  return 1;
}
